"""
This module handles the reading and processing of X-ray diffraction data.

Usage
`````
The usage of :mod:`~yadg.parsers.xrdtrace` can be specified by supplying 
``xrdtrace`` as an argument to the ``parser`` keyword of the `dataschema`. 
The parser supports the following parameters:

.. _yadg.parsers.xrdtrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.XRDTrace.Params

.. _yadg.parsers.xrdtrace.formats:

Formats
```````
The currently supported file formats are:

 - PANalytical ``xrdml`` files: :mod:`~yadg.parsers.xrdtrace.panalyticalxrdml`
 - PANalytical ``csv`` files: :mod:`~yadg.parsers.xrdtrace.panalyticalcsv`
 - PANalytical ``xy`` files: :mod:`~yadg.parsers.xrdtrace.panalyticalxy`

.. _yadg.parsers.xrdtrace.provides:

Provides
````````
:mod:`~yadg.parsers.xrdtrace` loads X-ray diffraction data, determines reasonable 
uncertainties of the signal intensity (y-axis), and explicitly populates the angle
axis (:math:`2\\theta`), if necessary. This raw data is stored, for each timestep, 
using the following format:

.. code-block:: yaml

    - raw:
        traces:
          "{{ trace_number }}":   # number of the trace
            angle:                
              {n: [!!float, ...], s: [!!float, ...], u: "deg"}
            intensity:
              {n: [!!float, ...], s: [!!float, ...], u: "counts"}

The uncertainties ``"s"`` in ``"angle"`` are taken as the step-width of
the linearly spaced :math:`2\\theta` values.

The uncertainties ``"s"`` of ``"intensity"`` are currently set to a constant
value of 1.0 count as all the supported files seem to produce integer values.

Metadata
````````
.. warning::
  
  The metadata collected from the raw data file depends on the filetype, and 
  is currently not strictly defined.

"""
from .main import process



"""
This module handles the reading and processing of X-ray photoelectron spectroscopy
data.

Usage
`````
The usage of :mod:`~yadg.parsers.xpstrace` can be specified by supplying
``xpstrace`` as an argument to the ``parser`` keyword of the `dataschema`.
The parser supports the following parameters:

.. _yadg.parsers.xpstrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.XPSTrace.Params

.. _yadg.parsers.xpstrace.formats:

Formats
```````
The currently supported file formats are:

 - ULVAC PHI Multipak XPS traces (``spe``) :mod:`~yadg.parsers.xpstrace.phispe`


.. _yadg.parsers.xpstrace.provides:

Provides
````````
:mod:`~yadg.parser.xpstrace` loads X-ray photoelectron spectroscopy data, 
determines uncertainties of the signal (y-axis), and explicitly populates the 
points in the energy axis (``E``). This raw data is stored, for each timestep, 
using the following format:

.. code-block:: yaml

    - raw:
        traces:
          "{{ trace_number }}":   # number of the trace
            name:          !!str  #
            atomic_number: !!str  #
            dwell_time:    !!str  #
            e_pass:        !!str  #
            description:   !!str  #
            E:                    # energies are always in eV
              {n: [!!float, ...], s: [!!float, ...], u: "eV"}
            y:                    # y-axis units from file
              {n: [!!float, ...], s: [!!float, ...], u: !!str}

The uncertainties ``"s"`` in ``"E"`` are taken as the step-width of
the linearly spaced energy values.

The uncertainties ``"s"`` of ``"y"`` are currently set to a constant
value of ``12.5`` counts per second as all the signals in the files seen so
far only seem to take on values in those steps.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/13

    Determining the uncertainty of the counts per second signal in XPS
    traces from the phispe parser should be done in a better way.

Metadata
````````
The metadata collected from the raw files depends on the ``filetype``. The 
following metadata entries are currently stored for each `step`:

.. code-block:: yaml

    params:
      software_id:  !!str  # software name
      version:      !!str  # software version
      username:     !!str  # operator

"""
from .main import process



"""
**tomatojson**: Processing of tomato electrochemistry outputs.
--------------------------------------------------------------

This module parses the electrochemistry ``json`` files generated by tomato.

.. warning::

    This parser is brand-new in `yadg-4.1` and the interface is unstable.

Four sections are expected in each tomato data file:

- ``technique`` section, describing the current technique,
- ``previous`` section, containing status information of the previous file,
- ``current`` section, containing status information of the current file,
- ``data`` section, containing the timesteps.

The reason why both ``previous`` and ``current`` are requires is that the device
status is recorded at the time of data polling, which means the values in ``current``
might be invalid (after the run has finished) or not in sync with the ``data`` (if
a technique change happened). However, ``previous`` may not be present in the first
data file of an experiment.

To determine the measurement errors, the values from BioLogic manual are used: for
measured voltages (:math:`E_{\\text{we}}` and :math:`E_{\\text{ce}}`) this corresponds
to a constant uncertainty of 0.004% of the applied E-range with a maximum of 75 uV, 
while for currents (:math:`I`) this is a constant uncertainty of 0.0015% of the applied
I-range with a maximum of 0.76 uA.

.. codeauthor:: Peter Kraus
"""

import json
import logging

logger = logging.getLogger(__name__)

I_ranges = {
    "1 A": 1e0,
    "100 mA": 1e-1,
    "10 mA": 1e-2,
    "1 mA": 1e-3,
    "100 uA": 1e-4,
    "10 uA": 1e-5,
    "1 uA": 1e-6,
    "100 pA": 1e-7,
}


def process(
    fn: str, encoding: str = "UTF-8", timezone: str = "UTC"
) -> tuple[list, dict, bool]:
    with open(fn, "r") as infile:
        jsdata = json.load(infile)

    technique = jsdata["technique"]
    previous = jsdata.get("previous", None)
    current = jsdata["current"]

    if "uts" in technique:
        uts = technique["uts"]
        fulldate = True
    else:
        uts = 0
        fulldate = False

    if previous is None:
        meta = current
    elif current["status"] == "STOP":
        meta = previous
    elif previous["elapsed_time"] > technique["start_time"]:
        meta = previous
    else:
        meta = current

    I_range = I_ranges[meta["I_range"]]
    E_range = meta["E_range"]["max"] - meta["E_range"]["min"]

    timesteps = []
    for point in jsdata["data"]:
        p = {"fn": fn, "uts": uts + technique["start_time"], "raw": {}}
        for k, v in point.items():
            if k == "time":
                p["uts"] += v
            elif k in {"Ewe", "Ece"}:
                s = max(E_range * 0.0015 / 100, 75e-6)
                p["raw"][k] = {"n": v, "s": s, "u": "V"}
            elif k in {"I"}:
                s = max(I_range * 0.004 / 100, 760e-12)
                p["raw"][k] = {"n": v, "s": s, "u": "A"}
            elif k in {"cycle"}:
                p["raw"]["cycle number"] = v
            else:
                logger.cricital(f"parameter {k}: {v} not understood.")
        p["raw"]["loop number"] = technique["loop_number"]
        p["raw"]["technique"] = technique["name"]
        p["raw"]["index"] = technique["index"]
        timesteps.append(p)

    return timesteps, None, fulldate



"""
The ``qftrace`` parser handles the reading and processing of the network analyzer 
traces, containing the reflection coefficient as a function of the sweeped frequency,
:math:`\\Gamma(f)`. The basic function of the parser is to:

1) read in the raw data and create timestamped `traces` 
2) detect the peaks in each trace (:math:`f_0`) and fit the quality factor :math:`Q_0`

Usage
`````
The use of :mod:`~yadg.parsers.qftrace` can be requested by supplying ``qftrace`` as
an argument to the ``parser`` keyword in the `dataschema`. The parser supports the
following parameters:

.. _yadg.parsers.qftrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.QFTrace.Params

.. _yadg.parsers.qftrace.formats:

 - LabView output in a tab-separated format (``csv``): 
   :mod:`~yadg.parsers.qftrace.labviewcsv`

.. _yadg.parsers.qftrace.provides:

Provides
````````
:mod:`~yadg.parsers.qftrace` loads the reflection trace data, determines the 
uncertainties of the signal (y-axis), and explicitly populates the points in 
the time axis (x-axis). This raw data is stored, for each
timestep, using the following format:

.. code-block:: yaml

  - raw:
      avg:             !!int  # number of scans that are averaged for each trace
      bw:                     # filter bandwith used to measure the trace
        {n: !!float, s: !!float, u: "Hz"}
      traces:
        "{{ trace_name }}":   # detector name, currently hard-coded to S11
          f:                  # frequency-axis units are always Hz
            {n: [!!float, ...], s: [!!float, ...], u: "Hz"} 
          Re(Γ):              # real part of the reflection coefficient
            {n: [!!float, ...], s: [!!float, ...], u: !!str}  
          Im(Γ):              # imaginary part of the reflection coefficient
            {n: [!!float, ...], s: [!!float, ...], u: !!str}  


The fitting of :math:`f_0` and :math:`Q_0` to all peaks found in each trace is 
performed by :mod:`~yadg.parsers.qftrace` automatically, and can be adjusted by 
specifying the ``method`` parameter and related options. The resulting data is 
stored in the ``derived`` entry in each `timestep`, and contains the following 
information:

.. code-block:: yaml

  - derived:
      "{{ trace_name }}":   # see above, currently set to S11
        f:                  # the frequencies of each peak
          {n: [!!float, ...], s: [!!float, ...], u: "Hz"} 
        Q:                  # the cavity quality factors for each peak
          {n: [!!float, ...], s: [!!float, ...], u: "Hz"} 
          
Metadata
````````
The metadata section is currently empty.

"""
from .main import process



from pydantic import BaseModel
from . import quadstarsac


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    parameters: BaseModel = None,
) -> tuple[list, dict, bool]:
    """Unified mass spectrometry data parser.

    This parser processes mass spectrometry scans in signal(mass) format.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg_dataschema.parameters.dataschema_4_1.MassTrace`.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag.

    """
    if parameters.filetype == "quadstar.sac":
        _data, _meta = quadstarsac.process(fn, encoding, timezone)
    return _data, _meta, True



"""
This parser handles the reading and processing of the legacy log files created by 
the LabView interface for the MCPT instrument. With a provided calibration, this 
parser calculates the temperature, inlet composition, and the inlet flow of the 
MCPT instrument.

.. warning::

    As of ``yadg-4.0.0``, this parser is deprecated and is no longer maintained.
    Please consider switching to other parsers.

Usage
`````
The use of :mod:`~yadg.parsers.meascsv` can be requested by supplying ``meascsv``
to the ``parser`` keyword in the `dataschema`. The following additional parameters
are supported by the parser:

.. _yadg.parsers.meascsv.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.MeasCSV.Params

.. _parsers_meascsv_provides:

Provides
````````
The parser is used to extract all of the tabular data in the input file, storing 
them using the column headers as keys. The functionality exposed by this parser
is using :mod:`~yadg.parsers.basiccsv` behind the scenes.

Metadata
````````
The metadata section is currently empty.


"""

from .main import process



from pydantic import BaseModel
import json
from ... import dgutils


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "localtime",
    parameters: BaseModel = None,
) -> tuple[list, dict, bool]:
    """
    A dummy parser.

    This parser simply returns the current time, the filename provided, and any
    ``kwargs`` passed.

    In case the provided ``filetype`` is a ``tomato.json`` file, this is a json
    data file from the :mod:`tomato` package, which should contain a :class:`list`
    of ``{"value": float, "time": float}`` datapoints in its ``data`` entry.

    Parameters
    ----------
    fn
        Filename to process

    encoding
        Not used.

    timezone
        Not used

    parameters
        Parameters for :class:`~dgbowl_schemas.yadg_dataschema.dataschema_4_1.parameters.Dummy`.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. No metadata is
        returned by the dummy parser. The full date is always returned.

    """
    if hasattr(parameters, "filetype") and parameters.filetype == "tomato.json":
        with open(fn, "r") as inf:
            jsdata = json.load(inf)
        ret = []
        for p in jsdata["data"]:
            ts = {
                "uts": p["time"],
                "fn": fn,
                "raw": {"value": {"n": p["value"], "s": 0.0, "u": " "}},
            }
            ret.append(ts)
        return ret, None, False
    else:
        kwargs = {} if parameters is None else parameters.dict()
        if "parser" in kwargs:
            del kwargs["parser"]
        result = {"uts": dgutils.now(), "fn": str(fn), "raw": kwargs}

        return [result], None, True



"""
This parser handles the reading and processing of any tabular files, as long as 
the first line contains the column headers, and the second line an optional
set of units. The columns of the table must be separated using a separator 
(``,`` or ``;`` or ``\\t`` or similar). 

A rudimentary column-converting functionality is also included. This allows the user 
to specify linear combinations of columns, and can be used to apply a calibration 
to the columnar data.

An attempt to deduce the timestamp from column headers is made automatically,
using :func:`yadg.dgutils.dateutils.infer_timestamp_from`. Alternatively, the 
timestamp column(s) and format can be provided using parameters.

Usage
`````
The use of :mod:`~yadg.parsers.basiccsv` can be requested by supplying
``basiccsv`` as an argument to the ``parser`` keyword of the `dataschema`.
The parser supports the following parameters:

.. _yadg.parsers.basiccsv.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.BasicCSV.Params

.. note::

    The specification of the calibration dictionary that ought to be passed via 
    ``convert`` (or stored as json in ``calfile``) is described in 
    :func:`~yadg.parsers.basiccsv.main.process_row`.

.. note::
    
    The ``calfile`` and ``convert`` functionalities allow for combining and 
    converting the raw data present in the data files into new entries, which
    are stored in the ``derived`` entry of each timestep.


Provides
````````
The primary functionality of :mod:`~yadg.parsers.basiccsv` is to load the tabular 
data, and determine the Unix timestamp. The headers of the tabular data are taken 
`verbatim` from the file, and appear as ``raw`` data keys.

Metadata
````````
The metadata section is currently empty.

"""

from .main import process



import logging
import json
import os
from typing import Union
from packaging import version
from dgbowl_schemas.yadg import to_dataschema, latest_version

from .. import dgutils
from .. import core

logger = logging.getLogger(__name__)


def calib_3to4(oldcal: dict, caltype: str) -> dict:
    newcal = {}
    if caltype == "calfile":
        for k, v in oldcal["detectors"].items():
            pd = {
                "window": (v.get("window", 3) - 1) // 2,
                "polyorder": v.get("poly", 2),
                "prominence": v.get("prominence", 1.0),
                "threshold": v.get("threshold", 1.0),
            }
            sp = {}
            for kk, vv in v["species"].items():
                if kk == "units" and vv == "min":
                    continue
                spec = {
                    "l": vv["l"] * 60.0,
                    "r": vv["r"] * 60.0,
                    "calib": {"inverse": {"slope": vv.get("rf", 1.0)}},
                }
                sp[kk] = spec
            id = {"det_1": 0, "det_2": 1}[v["id"]]
            newcal[k] = {"id": id, "peakdetect": pd, "species": sp}
    elif caltype == "Tcalfile":
        newcal = {"T": {"T_f": {"calib": {"linear": oldcal}}, "unit": "degC"}}
    elif caltype == "MFCcalfile":
        newcal = {}
        for k, v in oldcal.items():
            items = v.get("content", {k: 1.0})
            for kk, vv in items.items():
                if kk not in newcal:
                    newcal[kk] = {"unit": "ml/min"}
                newcal[kk][k] = {
                    "calib": {
                        "linear": {
                            "slope": v.get("slope", 1.0),
                            "intercept": v.get("intercept", 0.0),
                        }
                    },
                    "fraction": vv,
                }
    return newcal


def schema_3to4(oldschema: list) -> dict:
    newschema = {
        "metadata": {
            "provenance": {
                "type": "yadg update",
                "metadata": {
                    "yadg": dgutils.get_yadg_metadata(),
                    "update_schema": {"updater": "schema_3to4"},
                },
            },
            "version": latest_version,
            "timezone": "localtime",
        },
        "steps": [],
    }
    for oldstep in oldschema:
        newstep = {}

        newstep["parser"] = oldstep["datagram"]
        if newstep["parser"] == "gctrace":
            newstep["parser"] = "chromtrace"

        if "paths" in oldstep["import"]:
            oldstep["import"]["files"] = oldstep["import"].pop("paths")
        newstep["input"] = oldstep["import"]

        if oldstep.get("export", None) is not None:
            newstep["tag"] = oldstep["export"]

        calib = {}
        species = {}
        detectors = {}
        parameters = {}
        for k, v in oldstep["parameters"].items():
            if k in ["Tcalfile", "MFCcalfile", "calfile"]:
                try:
                    with open(v, "r") as cf:
                        temp = calib_3to4(json.load(cf), k)
                    logger.info(
                        "Found calfile '%s' at '%s', "
                        "will attempt to modernise the calibration.",
                        k,
                        v,
                    )
                    if k == "calfile":
                        for kk, vv in temp.items():
                            species[kk] = vv.pop("species")
                            detectors[kk] = vv
                    else:
                        calib.update(temp)
                except IOError:
                    logger.error(
                        "Error reading '%s' file: '{k}'. "
                        "Keeping original key-value pair.",
                        k,
                        v,
                    )
                    parameters[k] = v
            elif k == "method" and v == "q0refl":
                v = "kajfez"
                parameters[k] = v
            else:
                parameters[k] = v
        if newstep["parser"] == "meascsv" and "flow" not in calib:
            calib["flow"] = {"flow high": {}, "flow low": {}, "unit": "ml/min"}
        if calib != {}:
            parameters["convert"] = calib
        if species != {}:
            parameters["species"] = species
        if detectors != {}:
            parameters["detectors"] = detectors
        if parameters != {}:
            newstep["parameters"] = parameters
        newschema["steps"].append(newstep)

    return newschema


def update_object(objtype: str, object: Union[list, dict]) -> dict:
    """
    Yadg's update worker function.

    This is the main function called when **yadg** is executed as ``yadg update``.
    The main idea is to allow a simple update pathway from older versions of `schema` and
    ``datagram`` files to the current latest and greatest.

    Currently supports:

     - updating ``schema`` version 3.1.0 to 4.0.0
     - updating ``datagram`` version 3.1.0 to 4.0.0, although you **really** should be
       updating the ``schema`` instead and re-processing files.

    Parameters
    ----------
    objtype
        The type of the passed object, either `"datagram"` or `"schema"`.

    object
        The object to be updated

    Returns
    -------
    newobj: dict
        The updated and validated `"datagram"` or `"schema"`.

    """
    assert objtype in {"datagram", "schema"}, (
        f"Type '{objtype}' provided to update_object "
        "is not one of {'datagram', 'schema'}."
    )

    oldver = None
    if isinstance(object, list):
        oldver = "3.1.0"

    assert oldver is not None, (
        f"update_object: Impossible to figure out old '{type}' version. "
        "Please convert manually."
    )

    # distribute to updaters
    if version.parse(oldver).major == 3:
        assert (
            objtype == "schema"
        ), "Updating datagrams older than version 4.0.0 is not possible"
        logger.info("Updating old schema %s -> %s", oldver, latest_version)
        newobj = schema_3to4(object)
    elif version.parse(oldver).major == 4:
        logger.info("Already at latest version, no update necessary.")
        newobj = object

    with open("temp.json", "w") as outfile:
        json.dump(newobj, outfile, indent=1)

    if objtype == "schema":
        logger.info("Validating new schema.")
        newobj = to_dataschema(**newobj)
    elif objtype == "datagram":
        logger.info("Validating new datagram.")
        core.validators.validate_datagram(newobj)

    return newobj


def schema_from_preset(preset: dict, folder: str) -> dict:
    if isinstance(preset["metadata"]["provenance"], str):
        preset["metadata"]["provenance"] = "yadg preset"
    elif isinstance(preset["metadata"]["provenance"], dict):
        preset["metadata"]["provenance"] = {
            "type": "yadg preset",
            "metadata": {"preset_provenance": preset["metadata"]["provenance"]},
        }
    for step in preset["steps"]:
        inpk = "import" if "import" in step else "input"
        filk = "files" if "files" in step[inpk] else "folders"
        newf = []
        for oldf in step[inpk][filk]:
            if os.path.isabs(oldf):
                logger.warning(
                    "Item '%s' in '%s' is an absolute path and will not be patched.",
                    oldf,
                    filk,
                )
            else:
                assert not oldf.startswith("." + os.path.sep), (
                    f"Item '{oldf}' in '{filk}' does start with '.{os.path.sep}' and "
                    f"therefore should not be patched using '{folder}'."
                )
                newp = os.path.abspath(os.path.join(folder, oldf))
                newf.append(newp)
        step[inpk][filk] = newf
        if "calfile" in step["parameters"]:
            oldf = step["parameters"]["calfile"]
            if os.path.isabs(oldf):
                logger.warning(
                    "Specified calfile '%s' is an absolute path "
                    "and will not be patched.",
                    oldf,
                )
            else:
                newp = os.path.abspath(os.path.join(folder, oldf))
                step["parameters"]["calfile"] = newp
        if "externaldate" in step:
            using = "from" if "from" in step["externaldate"] else "using"
            if "file" in step["externaldate"][using]:
                oldf = step["externaldate"][using]["file"]["path"]
                if os.path.isabs(oldf):
                    logger.warning(
                        "Specified externaldate file '%s' is an absolute path "
                        "and will not be patched.",
                        oldf,
                    )
                else:
                    newp = os.path.abspath(os.path.join(folder, oldf))
                    step["externaldate"][using]["file"]["path"] = newp
    return preset



"""
The module handles the reading and processing of mass spectrometry files. The 
basic function of the parser is to:

1) read in the raw data and create timestamped `traces`
2) collect `metadata` such as the software version, author, etc.

Usage
`````
The use of :mod:`~yadg.parsers.masstrace` can be requested using the ``parser`` 
keyword in the `dataschema`. Further information can be specified in the
``"parameters"`` :class:`(dict)`:

.. _yadg.parsers.masstrace.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.MassTrace.Params

.. _yadg.parsers.masstrace.formats:

 - Pfeiffer Quadstar 32-bit scan analog data (``sac``): 
   :mod:`~yadg.parsers.masstrace.quadstarsac`
 
.. _yadg.parsers.masstrace.provides:

Provides
````````
The raw data, loaded from the supplied files, is stored using the following format:

.. code-block:: yaml

    - raw:
        traces:
          "{{ trace_number }}":  # number of the trace
            y_title: !!str       # y-axis label from file
            comment: !!str       # comment
            fsr:     !!str       # full scale range of the detector
            m/z:                 # masses are always in amu
              {n: [!!float, ...], s: [!!float, ...], u: "amu"}
            y:                   # y-axis units from file
              {n: [!!float, ...], s: [!!float, ...], u: !!str}

The uncertainties ``"s"`` in ``m/z`` are taken as the step-width of
the linearly spaced mass values.

The uncertainties ``"s"`` of ``y`` are the largest value between:

#. The quantization error from the ADC, its resolution assumed to be 32
   bit. Dividing F.S.R. by ``2 ** 32`` gives an error in the order of
   magnitude of the smallest data value in ``y``.
#. The contribution from neighboring masses. In the operating manual of
   the QMS 200 (see 2.8 QMS 200 F & 2.9 QMS 200 M), a maximum
   contribution from the neighboring mass of 50 ppm is noted.

.. note::

    The data in ``y`` may contain ``NaN`` s. The measured ion
    count/current value will occasionally exceed the specified detector
    F.S.R. (e.g. 1e-9), and will then flip directly to the maximum value
    of a float32. These values are set to ``float("NaN")``.

Metadata
````````
The metadata collected from the raw file will depend on the ``filetype``. 
In general, the following metadata entries are stored for each `step`:

.. code-block:: yaml

    params:
      software_id:  !!int  # software ID
      version:      !!str  # software version
      username:     !!str  # file author
 
"""
from .main import process



"""
panalyticalcsv: Processing of PANalytical XRD ``csv`` files
-----------------------------------------------------------

File Structure
``````````````

These files are split into a ``[Measurement conditions]`` and a ``[Scan points]`` 
section. The former stores the metadata and the latter all the datapoints. 

.. warning::
    
    This parser is fairly new and untested. As a result, the returned metadata 
    contain all the entries in the ``[Measurement conditions]`` section, without 
    any additional filtering.


Structure of Parsed Timesteps
`````````````````````````````

.. code-block:: yaml

    - fn:  !!str
    - uts: !!float
    - raw:
        traces:
          "{{ trace_number }}":  # Number of the trace.
            angle:               # Diffraction angle.
              {n: [!!float, ...], s: [!!float, ...], u: "deg"}
            intensity:           # Detector counts.
              {n: [!!float, ...], s: [!!float, ...], u: "counts"}

.. codeauthor:: Nicolas Vetsch
"""

from ...dgutils import dateutils
from .common import panalytical_comment, snake_case

# Converting camelCase xrdml keys to snake_case.


def _process_comments(comments: list[str]) -> dict:
    ret = {}
    for line in comments:
        ret.update(panalytical_comment(line))
    return ret


def _process_header(header: str) -> dict:
    """
    Processes the header section, staring with the ``[Measurement conditions]`` line.

    Parameters
    ----------
    header
        The header portion as a string.

    Returns
    -------
    header: dict
        A dictionary containing the processed metadata.

    """
    header_lines = header.split("\n")[1:-1]
    header = dict([l.split(",", 1) for l in header_lines])
    # Process comment entries.
    comments = []
    for key in list(header.keys()):
        if key.startswith("Comment"):
            comments.append(header.pop(key).strip('"'))
    comments = _process_comments(comments)
    # Renaming the keys.
    for key in list(header.keys()):
        header[snake_case(key)] = header.pop(key)
    header["comments"] = comments
    return header


def _process_data(data: str) -> tuple[list, list]:
    """
    Processes the data section, starting with the ``[Scan points]`` line.

    Parameters
    ----------
    data
        The data portion as a string.

    Returns
    -------
    (angle, intensity) : tuple[list, list]
        A list containing the angle data and one containing the
        intensity counts.

    """
    data_lines = data.split("\n")[1:-1]
    columns = data_lines[0].split(",")
    assert columns == ["Angle", "Intensity"], "Unexpected columns."
    datapoints = [l.split(",") for l in data_lines[1:]]
    angle, intensity = [list(d) for d in zip(*datapoints)]
    angle = [float(a) for a in angle]
    intensity = [float(i) for i in intensity]
    return angle, intensity


def process(
    fn: str, encoding: str = "utf-8", timezone: str = "UTC"
) -> tuple[list, dict, bool]:
    """
    Processes a PANalytical XRD csv file. All information contained in the header
    of the csv file is stored in the metadata.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "UTC".

    Returns
    -------
    (data, meta) : tuple[list, dict]
        (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and the full date tag.
        For .csv files tag is specified.

    """
    with open(fn, "r", encoding=encoding) as csv_file:
        csv = csv_file.read()
    # Split file into its sections.
    __, header, data = csv.split("[")
    assert header.startswith("Measurement conditions"), "Unexpected section."
    assert data.startswith("Scan points"), "Unexpected section."
    header = _process_header(header)
    # Process the data trace.
    angle, intensity = _process_data(data)
    angle = {
        "n": angle,
        "s": [float(header["scan_step_size"])] * len(angle),
        "u": "deg",
    }
    intensity = {
        "n": intensity,
        "s": [1.0] * len(intensity),
        "u": "counts",
    }
    # Process the metadata.
    uts = dateutils.str_to_uts(
        header["file_date_and_time"], format="%d/%B/%Y %H:%M", timezone=timezone
    )
    traces = {"0": {"angle": angle, "intensity": intensity}}
    data = [{"fn": fn, "uts": uts, "raw": {"traces": traces}}]
    return data, header, True



"""
This module handles the reading and processing of files containing electrochemical
data, including BioLogic's EC-Lab file formats.

.. note::

    This interface is not yet final and will change with version 5.0.0

Usage
`````
The usage of :mod:`~yadg.parsers.electrochem` can be specified by supplying
``electrochem`` as an argument to the ``parser`` keyword of the `dataschema`.
The parser supports the following parameters:

.. _yadg.parsers.electrochem.model:

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_4_1.step.ElectroChem.Params

.. _yadg.parsers.electrochem.formats:

Formats
```````
The currently supported file formats are:

 - EC-Lab raw data binary file and parameter settings
   :mod:`~yadg.parsers.electrochem.eclabmpr`
 - EC-Lab human-readable text export of data
   :mod:`~yadg.parsers.electrochem.eclabmpt`
 - tomato's structured json output
   :mod:`~yadg.parsers.electrochem.tomatojson`
  
Provides
````````
The basic function of the parser is to:

#. Read in the technique data and create timesteps.
#. Collect metadata, such as the measurement settings and the loops
   contained in a given file.
#. Collect data describing the technique parameter sequences.

.. note::

    ``.mpt`` files can contain more data than the corresponding binary
    ``.mpr`` file.

Most techniques write data that can be understood as a series of
timesteps. Each timestep provided by the parser has the following
format:

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        "{{ col1 }}":  !!int
        "{{ col2 }}":
          {n: !!float, s: !!float, u: !!str}

For impedance spectroscopy techniques (PEIS, GEIS), the data is made up
of spectroscopy traces. The data is thus split into traces by the column
``"cycle number"`` and each trace is cast into a single timestep. Each trace 
now corresponds to a spectroscopy scan, indexed by the technique name (PEIS or
GEIS). The timestep takes the following format:

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        traces:
          "{{ technique }}":
            "{{ col1 }}":
                [!!int, ...]
            "{{ col2 }}":
                {n: [!!float, ...], s: [!!float, ...], u: !!str}

.. note::

    The parsed data may contain infinities (i.e. ``float("inf")`` /
    ``float("-inf")``) or NaNs (i.e. ``float("nan")``). While datagrams 
    containing NaN and Inf can be exported and read back using python's json 
    module, they are not strictly valid jsons.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/10

    Current values of the uncertainties ``"s"`` are hard-coded from VMP-3 values
    of resolutions and accuracies, with ``math.ulp(n)`` as fallback. The values 
    should be device-specific, and the fallback should be eliminated.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/11

    The "raw" data in electrochemistry files should only contain the raw quantities,
    that is the ``control_I`` or ``control_V`` and the measured potentials ``Ewe``,
    ``Ece`` or the measured current ``I``. Analogous quantities should be recorded
    for PEIS/GEIS. All other columns should be computed by **yadg**.

Metadata
````````
The metadata collected from the raw file will depend on the ``filetype``. Currently,
no metadata is recorded for ``tomato.json`` filetype. For the ``eclab.mpt`` and 
``eclab.mpr`` filetypes, the metadata will contain a ``settings`` and a ``params`` 
field:

The ``settings`` field for parsed ``.mpt`` files contains the technique name, a 
posix timestamp and the raw header lines as found in the file. The ``settings`` 
from parsed ``.mpr`` files contain the technique and more explicitly parsed 
information than from ``.mpt`` files, like the "cell characteristics" specified 
in EC-Lab.

The ``params`` will contain the technique parameter sequences and the
keys in each sequence will be the same independent of ``filetype``, but
an :class:`int` value in the ``.mpr`` file may be a :class:`str` when 
parsed from the corresponding ``.mpt`` file, since the mapping has not 
yet been reverse engineered.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/12

    In ``.mpr`` files, some technique parameters in the settings module
    correspond to entries in drop-down lists in EC-Lab. These values are
    stored as single-byte values in ``.mpr`` files.

The metadata from parsed ``".mpr"`` files also provides the ``"log"``
which contains more general parameters, like software, firmware and
server versions, channel number, host address and an acquisition start
timestamp in Microsoft OLE format. 

.. note::

    If the ``.mpr`` file contains an ``ExtDev`` module (containing parameters
    of any external sensors plugged into the device), the ``log`` is usually 
    not present and therefore the full timestamp cannot be calculated.

"""
from .main import process
